# coding: utf-8

require "stringio"

# http://www.atmind.nl/blender/mystery_ot_blend.html

module IOExtLE
	def reade size
		r = read size
		r.size == size or raise "reached eof while reading #{size} bytes"
		r
	end
	def readw; reade(2).unpack("v")[0]; end
	def readd; reade(4).unpack("V")[0]; end
	def readq; reade(8).unpack("Q")[0]; end
	def readz
		r = ""
		loop {
			c = reade(1)
			c == "\0" and return r
			r += c
		}
	end
	def writed v; write [v].pack("V"); end
end

class IO; include IOExtLE; end
class StringIO; include IOExtLE; end

class Blend
	include Enumerable

	attr_reader :blocks, :types, :strcs, :ptrbytes

	class Block
		attr_accessor :code, :ptr, :sdna, :count
		attr_accessor :pos, :index, :type
		attr_reader :data
		def canoncode
			@code.gsub("\0", ".")
		end
		def size
			@data.size
		end
		def parsed
			return @parsed if @parsed

			@rem = nil
			if @type == nil
				@parsed = @data
				@rem = 0
			else
				StringIO.open(@data, "r") { |f|
					@parsed = @count.times.map { @type.read f }
					@rem = @data.size - f.pos
				}
			end
			@parsed
		end
		def data= d
			@data = d
			# invalidate
			@parsed = nil
			@rem = nil
		end
		def update!
			# note: update @data with invalidate by using self#data=.
			#       "self." before "data=" must be required? (or nothing happens??)
			self.data = StringIO.open("".b, "w") { |f| @parsed.map { |e| e.write f }; f.string }
		end
		def to_s
			"<Block:+%6X[%4d]%s (%d %s %d) %d/%d %sparsed%s>" % [
				@pos,
				@index,
				canoncode,
				@sdna,
				@type.name,
				@type.size,
				@count,
				@data.size,
				@rem == nil ? "NOT " : "",
				(@rem == nil or @rem == 0) ? "" :  " (remained #{@rem.inspect} bytes)"
				]
		end
		def inspect
			"#<Block +%X[%d]%s size=%X ptr=%X type=%s count=%d %sparsed%s>" % [
				@pos,
				@index,
				canoncode,
				@data.size,
				@ptr.pvalue,
				@type ? @type.name : "(none)",
				@count,
				@rem ? "" : "NOT",
				(!@rem || @rem == 0) ? "" : (" rem=%d" % @rem)
				]
		end
	end
	class Field
		attr_accessor :type, :name, :canonname
		def size
			@type.size
		end
	end
	class Type
		attr_accessor :index
	end
	class Prim < Type
		attr_accessor :name, :size
		def initialize name, size, index
			@name = name
			@size = size
			@index = index
		end
		def dump nest=0
			[" " * nest + @name]
		end
		def to_s
			"<T:#{@name} #{@size}>"
		end
		def read f
			PrimVal.new self, f.reade(@size)
		end
	end
	class Strc < Type
		attr_accessor :name, :sindex
		attr_reader :fields
		def fields= fs
			@fields = fs
			@size_cache = nil # invalidate
		end
		def size
			return @size_cache if @size_cache

			@size_cache = @fields.map{ |e|
				e.size
			}.inject(:+)
			@size_cache
		end
		def size=
			raise NotImplementedError("Strc size must not be changed without modifying fields!")
		end
		def dump nest=0
			@fields.map { |e| e.type.dump nest+1 }
		end
		def to_s
			"<S:#{@name} #{@fields.map{|e| (e.type.name || ??) + ?: + e.name.sub(e.canonname, ?/ + e.canonname + ?/) }.join ', '}>"
		end
		def read f
			r = {}
			@fields.each { |e|
				r[e.canonname] = e.type.read f
			}
			StrcVal.new self, r
		end
	end
	class Arr < Type
		attr_accessor :type, :count
		def name
			"[#{@count}]#{@type.name}"
		end
		def size
			@type.size * @count
		end
		def dump nest=0
			return [" " * nest + name]
		end
		def to_s
			"<A:#{@type} #{@count}>"
		end
		def read f
			# special handling: char[] is string primitive but its type is array of char.
			if @type.name == "char"
				return PrimVal.new self, f.read(@count)
			end
			ArrVal.new self, @count.times.map { @type.read f }
		end
	end
	class Ptr < Type
		attr_accessor :type, :size
		def name
			"*#{@type.name}"
		end
		def dump nest=0
			return [" " * nest + name]
		end
		def to_s
			"<P:#{@type}>"
		end
		def read f
			PtrVal.new self, f.reade(@size)
		end
	end
	class Func < Type
		attr_accessor :type
		def name
			"()#{@type.name}"
		end
		def size
			0 # dummy
		end
		def dump nest=0
			return [" " * nest + name]
		end
		def to_s
			"<F:#{@type}>"
		end
	end

	class Val
		attr_accessor :type, :value
		def initialize type, value
			@type = type
			@value = value
		end
		def to_s; inspect; end
		def insmul val, postfix=""
			val.inspect.lines.each_with_index.map { |l,i|
				if i == 0 # first line
					l.chomp + postfix + "\n"
				else
					"  " + l
				end
			}.join.chomp
		end
	end
	class PrimVal < Val
		def inspect
			if @type.is_a? Arr and @type.type == "char"
				return "@" + @value.inspect # expects String
			end
			case @type.name
			when "char"
				"%02X %4d %3d" % [@value.unpack("C")[0], @value.unpack("c")[0], @value.unpack("C")[0]]
			when "short"
				"%04X %6d %5d" % [@value.unpack("S")[0], @value.unpack("s")[0], @value.unpack("S")[0]]
			when "int"
				"%08X %11d %10d" % [@value.unpack("L")[0], @value.unpack("l")[0], @value.unpack("L")[0]]
			when "uint64_t"
				"%016X %21d %20d" % [@value.unpack("Q")[0], @value.unpack("q")[0], @value.unpack("Q")[0]]
			when "float"
				@value.unpack("f")[0].inspect
			else
				"(#{@type.name})" + @value.inspect
			end
		end
		def write f, path=""
			@type != nil and @value.size != @type.size and raise "?mismatch-value-size #{path} #{@value.size} expected #{@type.size}"
			f.write @value
		end
		def pvalue
			(@type.is_a? Arr and @type.type.name == "char") or raise "?pvalue-against-non-char-array #{@type}"
			@value.sub(/\0+$/, "")
		end
		def pvalue= val
			(@type.is_a? Arr and @type.type.name == "char") or raise "?pvalue=-against-non-char-array #{@type}"
			@value = val.ljust(@type.count, "\0")
		end
	end
	class StrcVal < Val
		def inspect
			r = "strc %s\n" % @type.name
			off = 0
			@type.fields.each { |e|
				val = @value[e.canonname]
				valstr = insmul val, " ; +%X" % off
				r += "  #{e.type.name} #{e.name}: #{valstr}\n"
				off += val.type ? val.type.size : val.value.size
			}
			r
		end
		def write f, path=""
			k = @type.fields.any? { |e| not @value.key? e.canonname } and raise "?warning-unknown-key #{path}.#{k}"
			@type.fields.each { |e|
				@value[e.canonname].write f, path + "." + e.name
			}
		end
	end
	class ArrVal < Val
		def inspect
			r = "[" + @value.map{ |e| e.inspect }.join(", ") + "]"
			if 80 < r.size or r.include? "\n" or @value.find { |e| e.is_a? StrcVal }
				r = "[\n"
				@value.each { |e|
					r += "  " + insmul(e) + ",\n"
				}
				r += "]\n"
			end
			r
		end
		def write f, path=""
			@value.size != @type.count and raise "?mismatch-value-count #{path} #{@value.size} expected #{@type.count}"
			@value.each_with_index { |e,i|
				e.write f, path + "[#{i}]"
			}
		end
	end
	class PtrVal < Val
		def inspect
			("*%%0%dX" % (@type.size * 2)) % @value.unpack(@type.size == 4 ? "V" : "Q")[0]
		end
		def write f, path=""
			@value.size != @type.size and raise "?mismatch-value-size #{path} #{@value.size} expected #{@type.size}"
			f.write @value
		end
		def pvalue
			@value.unpack(@type.size == 4 ? "V" : "Q")[0]
		end
		def pvalue= ptr
			@value = [ptr].pack(@type.size == 4 ? "V" : "Q")
		end
	end

	def initialize f
		f.read(7) == "BLENDER" or raise "?signat"
		bitwidth = f.read(1)
		case bitwidth
		when "_" # 32bit
			@ptrbytes = 4
		when "-" # 64bit
			@ptrbytes = 8
		else
			raise "?bitwidth #{bitwidth.inspect}"
		end
		f.read(1) == "v" or raise "?LE"
		@ver = f.read(3)
		puts "ver: %s" % @ver
		@blocks = []
		index = 0
		tvoid = Ptr.new
		tvoid.type = Prim.new "void", 0, nil
		tvoid.size = @ptrbytes
		while not f.eof
			block = Block.new
			block.pos = f.pos
			block.index = index
			index += 1
			block.code = f.read(4)
			size = f.readd
			block.ptr = PtrVal.new tvoid, f.read(@ptrbytes)
			block.sdna = f.readd
			block.count = f.readd
			block.data = f.read(size)
			@blocks << block
		end

		parse_dna1
		parse_blocks

		self
	end

	def self.read fn
		open(fn, "rb") { |f| new f }
	end

	def parse_dna1
		dna1 = @blocks.find { |e| e.code == "DNA1" }

		puts "parsing DNA1"

		names, xtypes, strcs = nil # make local-vars out of following block

		StringIO.open(dna1.data, "r") { |f|
			f.read(4) == "SDNA" or raise "?signat-dna1"

			f.read(4) == "NAME" or raise "?signat-dna1-name"
			nnames = f.readd
			names = nnames.times.map { f.readz }

			f.pos = f.pos + (4 - (f.pos % 4)) % 4 # TODO check align is for block data, or file.
			f.read(4) == "TYPE" or raise "?signat-dna1-type %X+18+%X" % [dna1.pos, f.pos - 4]
			ntypes = f.readd
			types = ntypes.times.map { f.readz }

			f.pos = f.pos + (4 - (f.pos % 4)) % 4 # TODO check align is for block data, or file.
			f.read(4) == "TLEN" or raise "?signat-dna1-tlen %X+18+%X" % [dna1.pos, f.pos - 4]
			tlens = ntypes.times.map { f.readw }

			puts "pass0 create #{ntypes} types as primitive"
			@types = ntypes.times.map { |i| Prim.new types[i], tlens[i], i }

			f.pos = f.pos + (4 - (f.pos % 4)) % 4 # TODO check align is for block data, or file.
			f.read(4) == "STRC" or raise "?signat-dna1-strc %X+18+%X" % [dna1.pos, f.pos - 4]
			xtypes = @types.dup
			nstrcs = f.readd
			puts "pass1 load #{nstrcs} structures"
			@strcs = nstrcs.times.map { |i|
				strc = Strc.new
				itype = f.readw
				type = @types[itype]
				strc.name = type.name
				strc.index = type.index
				strc.sindex = i
				strc.fields = f.readw.times.map {
					field = Field.new
					iftype = f.readw
					field.type = @types[iftype] or raise "?oob-type #{iftype}/#{@types.size}"
					ifname = f.readw
					field.name = names[ifname] or raise "?oob-name #{ifname}/#{names.size}"
					field
				}
				xtypes[itype] = strc
				strc
			}
		}

		# first (temporary) index
		xklasses = {}
		xtypes.each { |t| xklasses[t.name] = t }

		# migrate fields' type
		puts "pass2 transform structure fields type"
		@strcs.each { |e|
			e.fields.each { |f|
				f.type.name == xtypes[f.type.index].name or raise "?discrepancy-type-xtype #{f.type.index} #{f.type.name} != #{xtypes[f.type.index].name}"
				f.type = xtypes[f.type.index] or raise "?not-found-for-index #{f.type}"

				# note: first ptr then array. because *foo[3] is not "ptr to arr of foo", it is "arr of ptr to foo"
				canonname = f.name
				if canonname =~ /\((.+)\)\(\)/
					canonname = $1
					t = Func.new
					t.type = f.type
					f.type = t
				end
				while canonname =~ /^\*/
					canonname = canonname[1..-1]
					t = Ptr.new
					t.type = f.type
					t.size = @ptrbytes
					f.type = t
				end
				while canonname =~ /(.+)(?:\[(\d+)\])$/
					canonname = $1
					t = Arr.new
					t.type = f.type
					t.count = $2.to_i
					f.type = t
				end

				f.canonname = canonname
			}
		}

		# migrate types with size check
		puts "pass3 migrate structures as types"
		@strcs.each { |e|
			if @types[e.index].size != e.size
				raise "?size-mismatch: #{e.name} TYPE=#{@types[e.index].size} STRC=#{e.size}"
			end
			@types[e.index] = e
		}

		puts "pass4 profit!"
		# second (final) index: not required, for convenient
		#@klasses = {}
		#@types.each { |t| @klasses[t.name] = t }
	end

	def parse_blocks
		puts "parsing blocks"
		@byptr = {}
		@blocks.each { |b|
			b.type = b.sdna == 0 ? nil : @strcs[b.sdna] # note: not @types!
			#b.parsed
			@byptr.key? b.ptr.value and raise "?conflict-ptr #{b.ptr} #{byptr[b.ptr.value]} #{b}"
			@byptr[b.ptr.value] = b
		}
		puts "ok"
	end

	def to_s
		inspect
	end
	def inspect
		"#<Blend nblocks=%d ntypes=%d nstrcs=%d>" % [@blocks.size, @types.size, @strcs.size]
	end
	def [] i
		@blocks[i]
	end
	def each &block # note: &block is important! without it, each{...} and Enumerable methods totally don't work.
		@blocks.each &block
	end
	def bypos pos
		@blocks.find { |e| e.pos == pos }
	end
	def byptr ptr
		@byptr[ptr]
	end
	def write f
		f.write "BLENDER"
		f.write @ptrbytes == 4 ? "_" : "-"
		f.write "v" # LE
		f.write @ver

		@blocks.each { |e|
			f.write e.code
			f.writed e.data.size
			e.ptr.write f
			f.writed e.sdna
			f.writed e.count
			f.write e.data
		}
	end
end

if $0 == __FILE__
	case ARGV.shift
	when "dump"
		blend = open(ARGV[0], "rb") { |f| Blend.new f }
		blend.blocks.each_with_index { |b,i|
			puts "block %s[%d]+%X %d/%d %s" % [b.canoncode, i, b.pos, b.size, b.count, b.ptr.inspect]
			if b.type == nil
				puts b.parsed.inspect
			elsif b.count == 1
				puts b.parsed[0].inspect
			else
				b.parsed.each_with_index { |e,i|
					puts ("[%d] " % i) + e.inspect
				}
			end
		}
	when "diet"
		blend = open(ARGV[0], "rb") { |f| Blend.new f }

		# ScrArea.spacebase.first/last = 0 ; remove all area's views' info except current view.
		blend.select { |e|
			e.type and
			e.type.name == "ScrArea" and
			e.parsed[0].value["spacedata"].value["first"].pvalue != 0
		}.each { |e|
			spacedata = e.parsed[0].value["spacedata"].value
			pvview = spacedata["first"]
			spacedata["last"].value = pvview.value # note: .value=.value is faster than .pvalue=.pvalue
			e.update!
			view = blend.byptr pvview.value
			view or next # note: sometimes nil...
			view.parsed[0].value["next"].pvalue = 0
			view.parsed[0].value["prev"].pvalue == 0 or raise "?first-view-but-prev-exists"
			view.update!
		}

		# ARegion.panels.first/last = 0 ; resetting panel open/close/order
		blend.select { |e|
			e.type and
			e.type.name == "ARegion" and
			e.parsed[0].value["panels"].value["first"].pvalue != 0
		}.each { |e|
			e.parsed[0].value["panels"].value["first"].pvalue = 0
			e.parsed[0].value["panels"].value["last"].pvalue = 0
			e.update!
		}

		# SpaceOops.treestore = 0 ; resetting Outliner's? tree state
		blend.select { |e|
			e.type and
			e.type.name == "SpaceOops"
		}.each { |e|
			e.parsed[0].value["treestore"].pvalue = 0
			e.update!
		}

		ofn = ARGV[1] || ARGV[0].sub(/\.blend$/, "") + "_diet.blend"
		open(ARGV[0] + ".tmp", "wb") { |f| blend.write f }
		File.rename ARGV[0] + ".tmp", ofn
	when "sniff"
		blend = open(ARGV[0], "rb") { |f| Blend.new f }

		erase = { f: [], t: [], s: [], c: nil }
		if ARGV[1]
			ARGV[2..-1].each { |e|
				type = e[0]
				arg = e[1..-1]
				if arg == ""
					if type == "-"
						erase = {}
					else
						erase[type.to_sym] ||= []
					end
				elsif type == "-"
					arg.chars.each { |c|
						erase[c.to_sym] = nil
					}
				else
					evalue = nil
					if arg =~ /(.+)=(.*)$/
						evalue = $2
						arg = $1
					end
					case arg
					when /^(\d+)$/
						erase[type.to_sym] << [$1.to_i, evalue]
					when /^(\d+)..(\d+)$/
						erase[type.to_sym] << [($1.to_i)..($2.to_i), evalue]
					else
						raise "?filter-syntax-error"
					end
				end
			}
		end

		updated = false

		def walkprim val, path, &block
			case val
			when Blend::PrimVal
				yield val, path
			when Blend::StrcVal
				val.type.fields.each { |e|
					walkprim val.value[e.canonname], path + "." + e.name, &block
				}
			when Blend::ArrVal
				val.value.each_with_index { |e,i|
					walkprim e, path + "[#{i}]", &block
				}
			when Blend::PtrVal
				# nothing
			else
				raise "?unknown-val #{path} -> #{val.class}"
			end
		end

		# filename-likes [256][1024][1056][1090]
		if erase[:f]
			blend.each { |blk|
				blk.type == nil and next # skip blob blocks
				bup = false
				blk.parsed.each_with_index { |ee,i|
					walkprim(ee, "##{blk.index}[#{i}]") { |val, path|
						if val.type.is_a? Blend::Arr and val.type.type.name == "char" and [256, 1024, 1056, 1090].include? val.type.count and val.pvalue != ""
							puts "filename-looks at #{path} -> #{val.pvalue.inspect}"
							if e = erase[:f].find { |e| e[0] === blk.index }
								val.pvalue = e[1] || ""
								puts " ...erased"
								bup = true
								updated = true
							end
						end
					}
				}
				bup and blk.update!
			}
		end

		# text original-filename
		if erase[:t]
			blend.select{|blk| blk.type and blk.parsed[0].type.name == "Text" and blk.parsed[0].value["name"].pvalue != 0 }.each { |blk|
				tobj = blend.byptr(blk.parsed[0].value["name"].value)
				puts "text filename at ##{blk.index} -> ##{tobj.index} -> #{tobj.data.inspect}"
				if erase[:t].any? { |e| e[0] === blk.index }
					blk.parsed[0].value["name"].pvalue = 0 # TODO is this ok? note: ignoreing e[1]
					puts " ...erased"
					blk.update!
					updated = true
				end
			}
		end

		# slack of char[] (=non-\0s after first \0)
		if erase[:s]
			blend.each { |blk|
				blk.type == nil and next # skip blob blocks
				bup = false
				blk.parsed.each_with_index { |ee,i|
					walkprim(ee, "##{blk.index}[#{i}]") { |val, path|
						if val.type.is_a? Blend::Arr and val.type.type.name == "char" and val.value =~ /^[\x20-\x7e]*\0[\x20-\x7e]/
							puts "slack at #{path} -> #{val.pvalue.inspect}"
							if erase[:s].any? { |e| e[0] === blk.index }
								val.pvalue = val.pvalue.sub(/^([\x20-\x7e]*)\0.*/) { $1 } # note: ignoreing e[1]
								puts " ...erased"
								bup = true
								updated = true
							end
						end
					}
				}
				bup and blk.update!
			}
		end

		# name of customdata
		if erase[:c]
			blend.select{|blk| blk.type and blk.parsed[0].type.name == "CustomDataLayer" }.each { |blk|
				bup = false
				blk.parsed.each_with_index { |ee,i|
					if ee.type.name == "CustomDataLayer" and ee.value["name"].pvalue != ""
						puts "CustomDataLayer name at ##{blk.index}[#{i}] -> #{ee.value["name"].pvalue}"
						if e = erase[:c].find { |e| e[0] === blk.index }
							ee.value["name"].pvalue = e[1] || ""
							puts " ...erased"
							bup = true
							updated = true
						end
					end
				}
				bup and blk.update!
			}
		end

		if ARGV[1] and updated
			ofn = ARGV[1] || ARGV[0].sub(/\.blend$/, "") + "_clean.blend"
			open(ARGV[0] + ".tmp", "wb") { |f| blend.write f }
			File.rename ARGV[0] + ".tmp", ofn
			puts "blend file wrote."
		end
	else
		puts "?unknown-command"
	end
end
