#include <libdragon.h>
#include <stdint.h>

#include <PR/mbi.h>
#include <PR/sptask.h>
#include <PR/ucode.h>

#include "../../include/regsinternal.h"

#include "rsp.h"

static volatile struct SP_regs_s * const SP_regs = (struct SP_regs_s *)0xa4040000;
static volatile u32 * const SP_PC = (u32*)0xa4080000;

void waittask() {
	while((SP_regs->status & 0x3) == 0) /*nothing*/; /* wait until SP stops (halt or broke) */
}

void runtask(OSTask *tsk) {
	// stop RSP and wait its stop
	SP_regs->status = 0x2B2E; /* set halt,clear broke,intr,sstep,,SetIoB,clear sig0,1,2 */
	waittask();
	*SP_PC = 0x000;

#define PHYSICAL(p) ((uintptr_t)(p) & 0x3FFFffff)
#define MAKE_PHYSICAL(p) ((p) = (u64*)PHYSICAL(p))

	// load ucode_boot
	// note: tsk->t.ucode_boot itself is not stored as converted in sptask.o.
	SP_regs->RSP_addr = (void*)0x1000;
	SP_regs->RDRAM_addr = (void*)PHYSICAL(tsk->t.ucode_boot);
	MEMORY_BARRIER();
	SP_regs->rsp_read_length = ((tsk->t.ucode_boot_size + 0x7) & 0xFFFFfff8) - 1;
	MEMORY_BARRIER();
	while(SP_regs->status & 0x1C) /*nothing*/; /* wait until DMA/IO is not busy/full */
	MEMORY_BARRIER();

	// convert OSTask
	MAKE_PHYSICAL(tsk->t.ucode);
	MAKE_PHYSICAL(tsk->t.ucode_data);
	MAKE_PHYSICAL(tsk->t.dram_stack);
	MAKE_PHYSICAL(tsk->t.output_buff);
	MAKE_PHYSICAL(tsk->t.output_buff_size);
	MAKE_PHYSICAL(tsk->t.data_ptr);
	MAKE_PHYSICAL(tsk->t.yield_data_ptr);

	// load OSTask
	SP_regs->RSP_addr = (void*)0x0FC0;
	SP_regs->RDRAM_addr = (void*)PHYSICAL(tsk);
	MEMORY_BARRIER();
	SP_regs->rsp_read_length = 0x03F;
	MEMORY_BARRIER();
	while(SP_regs->status & 0x1C) /*nothing*/; /* wait until DMA/IO is not busy/full */
	MEMORY_BARRIER();

	// go
	SP_regs->status = 0x0001; /* clear halt */
}

void init_gfx_task(OSTask *tsk, u64 *dram_stack, u64 *output_buf, Gfx *glist, int debug) {
	OSTask_t *t = &tsk->t;

	t->type = M_GFXTASK;
	t->flags = OS_TASK_DP_WAIT;

	t->ucode_boot = (u64*)rspbootTextStart;
	// note: char* is important! ptr minus ptr is not bytes but elements!
	t->ucode_boot_size = (char*)rspbootTextEnd - (char*)rspbootTextStart;

	t->ucode = (u64*)gspF3DEX2_fifoTextStart;
	t->ucode_size = SP_UCODE_SIZE;

	t->ucode_data = (u64*)gspF3DEX2_fifoDataStart;//my_ucode_data;
	t->ucode_data_size = SP_UCODE_DATA_SIZE;

	t->dram_stack = dram_stack;
	t->dram_stack_size = SP_DRAM_STACK_SIZE8;

	t->output_buff = output_buf;
	t->output_buff_size = t->output_buff + (0x1000/*RDP_OUTPUT_LEN?*/ / sizeof(u64));

	t->data_ptr = (u64*)glist;
	t->data_size = debug; /* ignored */

	t->yield_data_ptr = (u64*)0xc0800000;
	t->yield_data_size = 0;
}

void init_dummy_task(OSTask *tsk, u32 type) {
	OSTask_t *t = &tsk->t;
	// note: ucodes are DMA'd, so must be aligned to 8 bytes, and at lest 8 bytes length and multiple by 8.
	static const u32 const ucode_break[] __attribute__((aligned(8))) = { 0x0000000D, 0x00000000 };

	// must be filled for HLE emulators.
	t->type = type;

	// used by runtask().
	t->ucode_boot = (u64*)ucode_break;
	t->ucode_boot_size = sizeof(ucode_break);

	// these are dummy. usually used by rspboot but there are no rspboot, so we can totally ignore these.
	t->flags = 0;
	t->ucode = (u64*)0;
	t->ucode_size = 0;
	t->ucode_data = (u64*)0;
	t->ucode_data_size = 0;
	t->dram_stack = (u64*)0;
	t->dram_stack_size = 0;
	t->output_buff = (u64*)0;
	t->output_buff_size = (u64*)0;
	t->data_ptr = (u64*)0;
	t->data_size = 0;
	t->yield_data_ptr = (u64*)0;
	t->yield_data_size = 0;
}

void init_rspboot_dummy_task(OSTask *tsk, u32 type) {
	OSTask_t *t = &tsk->t;
	// note: ucodes and its datas are DMA'd, so must be aligned to 8 bytes, and at lest 8 bytes length and multiple by 8.
	// note2: rspboot always transfer 0xF80 bytes, so we must prepare 0xF80 bytes.
	static const u32 const ucode_break[0xF80 / sizeof(u32)] __attribute__((aligned(8))) = { 0x0000000D, 0x00000000, };
	// note: ucode_data < 8bytes are not allowed by rspboot!
	static u32 ucode_data[] __attribute__((aligned(8))) = { 0x00000000, 0x00000000 };

	// must be filled for HLE emulators.
	t->type = type;

	// used by runtask().
	t->ucode_boot = (u64*)rspbootTextStart;
	// note: char* is important! ptr minus ptr is not bytes but elements!
	t->ucode_boot_size = (char*)rspbootTextEnd - (char*)rspbootTextStart;

	// those are used by rspboot.
	t->flags = 0; // note: only OS_TASK_DP_WAIT is checked and used.
	t->ucode = (u64*)ucode_break;
	t->ucode_size = sizeof(ucode_break); // note: ignored by rspboot. (see ucode_break comment)
	t->ucode_data = (u64*)ucode_data;
	t->ucode_data_size = sizeof(ucode_data); // note: 0-byte is not allowed!

	// those are for ucode, not rspboot. fill dummy.
	t->dram_stack = (u64*)0;
	t->dram_stack_size = 0;
	t->output_buff = (u64*)0;
	t->output_buff_size = (u64*)0;
	t->data_ptr = (u64*)0;
	t->data_size = 0;
	t->yield_data_ptr = (u64*)0;
	t->yield_data_size = 0;
}

