#include <libdragon.h>

#include "adler32.h"

#include "../../include/regsinternal.h"

/* FIXME this should be in include/regsinternals.h */
typedef struct PI_regs_ex_s {
	PI_regs_t r;

	uint32_t dom1_latency;
	uint32_t dom1_pulse;
	uint32_t dom1_pagesize;
	uint32_t dom1_rel;
	uint32_t dom2_latency;
	uint32_t dom2_pulse;
	uint32_t dom2_pagesize;
	uint32_t dom2_rel;
} PI_regs_ex_t;

/* const: PI reg and others that related to SRAM */
static volatile struct PI_regs_ex_s * const PI_regs = (struct PI_regs_ex_s*)0xA4600000;
static const uint32_t SRAM_BASE = 0x08000000;

static uint32_t g_pagesize /* = 0*/;
static uint32_t g_pagecopy /* = 0*/;
static uint32_t g_save_safety /* = 0*/;

/* FIXME dma_raw_read and dma_raw_write should be in libdragon/dma.c */

static void dma_raw_read(void * ram_address, unsigned long pi_address, unsigned long len) {
	disable_interrupts();

	while (dma_busy()) ;
	MEMORY_BARRIER();
	PI_regs->r.ram_address = ram_address;
	PI_regs->r.pi_address = pi_address & 0x1FFFFFFF;
	MEMORY_BARRIER();
	PI_regs->r.write_length = len - 1; /* PI *writes* to RDRAM */
	MEMORY_BARRIER();

	enable_interrupts();

	while (dma_busy()) ;
}

static void dma_raw_write(void * ram_address, unsigned long pi_address, unsigned long len) {
	disable_interrupts();

	while (dma_busy()) ;
	MEMORY_BARRIER();
	PI_regs->r.ram_address = ram_address;
	PI_regs->r.pi_address = pi_address & 0x1FFFFFFF;
	MEMORY_BARRIER();
	PI_regs->r.read_length = len - 1; /* PI *reads* from RDRAM */
	MEMORY_BARRIER();

	enable_interrupts();

	while (dma_busy()) ;
}

void save_init() {
	PI_regs->dom2_latency = 5;
	PI_regs->dom2_pulse = 12;
	PI_regs->dom2_pagesize = 13;
	PI_regs->dom2_rel = 2;
}

/* note: checksum is written at last 4 bytes, so usable area is pagesize-4 bytes length. */
/* note2: pagesize are aligned by 8bytes. */
void save_pagesize(uint32_t pagesize) {
	if(pagesize < 8) {
		// too small, at least 4 bytes for checksum, and align 8.
		// maximum these are 8.
		g_save_safety = 0;
		return;
	}
	if(0x8000 < pagesize * g_pagecopy) {
		// too large, 32k is exist, don't allow larger than that.
		// if you want to set pagesize=32k, set pagecopy=1 first.
		g_save_safety = 0;
		return;
	}
	if((pagesize & 0x00000007) != 0) {
		// size is not aligned 8bytes.
		g_save_safety = 0;
		return;
	}
	g_pagesize = pagesize;
	g_save_safety = 1;
}

void save_pagecopy(uint32_t pagecopy) {
	if(pagecopy < 1) {
		// illegal!
		g_save_safety = 0;
		return;
	}
	if(0x8000 < g_pagesize * pagecopy) {
		// too large, 32k is exist, don't allow larger than that.
		// if you want to set 3 or more, set pagesize <= 0x8000/2 first.
		g_save_safety = 0;
		return;
	}
	g_pagecopy = pagecopy;
	g_save_safety = 1;
}

/* uint32_t should be bool. */
/* note: content of buf is destroyed even when read fails (i.e. bad checksum). */
uint32_t save_read(void *buf) {
	if(!g_save_safety || ((uintptr_t)buf & 0x00000007) != 0) {
		return 0; // Unsafe!!
	}

	for(int i = 0; i < g_pagecopy; i++) {
		dma_raw_read((void*)((uintptr_t)buf & 0x007Fffff), SRAM_BASE + g_pagesize * i, g_pagesize);
		uint32_t dsum = *(uint32_t*)((char*)buf + g_pagesize - 4);
		uint32_t csum = adler32(buf, g_pagesize - 4);
		if(dsum == csum) {
			return 1; // OK!
		}
	}
	return 0; // Failed...
}

/* uint32_t should be bool. */
/* note: last 4 bytes of buf are overwritten by checksum. */
/* note2: save_write does not verify, it requires additional buffer. */
uint32_t save_write(void *buf) {
	if(!g_save_safety || ((uintptr_t)buf & 0x00000007) != 0) {
		return 0; // Unsafe!!
	}

	*(uint32_t*)((char*)buf + g_pagesize - 4) = adler32(buf, g_pagesize - 4);
	for(int i = 0; i < g_pagecopy; i++) {
		dma_raw_write((void*)((uintptr_t)buf & 0x007Fffff), SRAM_BASE + g_pagesize * i, g_pagesize);
	}

	return 1;
}
